SystemOrganization addCategory: #Git!
SystemOrganization addCategory: #'Git-Plumbing'!
SystemOrganization addCategory: #'Git-Tests'!

Object subclass: #GitConfig
	instanceVariableNames: 'entries'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitConfig methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 15:59'!
category: aString
	^ entries at: aString ifAbsent: [Dictionary new]! !

!GitConfig methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 16:01'!
configAt: key
	| parts |
	parts := key findTokens: '.'.
	^ (self category: parts first) at: parts second ifAbsent: [nil]! !

!GitConfig methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 15:59'!
entries
	^ entries! !

!GitConfig methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 15:56'!
initialize
	super initialize.
	entries := Dictionary new.! !

!GitConfig methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 15:58'!
readFrom: aStream
	| category |
	category := nil.
	aStream contents linesDo: [:line |
		line := line withBlanksTrimmed.
		line isEmpty ifFalse: [
			line first = $[
				ifTrue: [
					category := entries at: (line copyFrom: 2 to: line size - 1) ifAbsentPut: [Dictionary new]]
				ifFalse: [ | pos |
					pos := line indexOf: $=.
					pos = 0 ifFalse: [
						category
							at: (line copyFrom: 1 to: pos - 1) withBlanksTrimmed
							put: (line copyFrom: pos + 1 to: line size) withBlanksTrimmed]]]]! !

Object subclass: #GitDeltaInterpreter
	instanceVariableNames: 'source instructions output commands'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitDeltaInterpreter methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 12:01'!
instructions
	^ instructions! !

!GitDeltaInterpreter methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 12:01'!
instructions: anObject
	instructions := anObject! !

!GitDeltaInterpreter methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 12:09'!
readVariableInteger
	| val shift b |
	val := 0.
	shift := 0.
	[
		b := commands next.
		val := val bitOr: ((b bitAnd: 127) bitShift: shift).
		shift := shift + 7.
	] doWhileFalse: [(b bitAnd: 128) isZero].
	^ val! !

!GitDeltaInterpreter methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 12:20'!
run
	[commands atEnd] whileFalse: [self runNextCommand]! !

!GitDeltaInterpreter methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 12:25'!
runAgainst: aByteArray
	source := aByteArray.
	
	commands := ReadStream on: instructions.
	source size = self readVariableInteger ifFalse: [self error: 'Delta source size mismatch'].
	output := WriteStream on: (ByteArray new: self readVariableInteger).
	self run.

	^ output contents! !

!GitDeltaInterpreter methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 12:19'!
runNextCommand
	| cmd |
	cmd := commands next.
	cmd = 0 ifTrue: [self error: 'unexpected delta opcode 0'].
	cmd < 128
		ifTrue: [output nextPutAll: (commands next: cmd)]
		ifFalse: [
			| off size |

			off := 0.
			(cmd bitAnd: 2r00000001) isZero ifFalse: [off := commands next].
			(cmd bitAnd: 2r00000010) isZero ifFalse: [off := off bitOr: (commands next bitShift: 8)].
			(cmd bitAnd: 2r00000100) isZero ifFalse: [off := off bitOr: (commands next bitShift: 16)].
			(cmd bitAnd: 2r00001000) isZero ifFalse: [off := off bitOr: (commands next bitShift: 24)].

			size := 0.
			(cmd bitAnd: 2r00010000) isZero ifFalse: [size := commands next].
			(cmd bitAnd: 2r00100000) isZero ifFalse: [size := size bitOr: (commands next bitShift: 8)].
			(cmd bitAnd: 2r01000000) isZero ifFalse: [size := size bitOr: (commands next bitShift: 16)].
			size = 0 ifTrue: [size := 16r10000].
			
			output nextPutAll: (source copyFrom: off + 1 to: off + size)]! !

Object subclass: #GitObject
	instanceVariableNames: 'objectName'
	classVariableNames: 'TypecodeMap'
	poolDictionaries: ''
	category: 'Git-Plumbing'!

GitObject subclass: #GitBlob
	instanceVariableNames: 'bytes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitBlob class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 16:02'!
bytes: aByteArray
	^ self new initializeBytes: aByteArray asByteArray! !

!GitBlob class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 07:19'!
typeCode
	^ 'blob'! !

!GitBlob methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 07:59'!
bytes
	^ bytes! !

!GitBlob methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 13:31'!
initializeBytes: aByteArray
	bytes := aByteArray! !

!GitBlob methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:49'!
loadFrom: aStream name: anObjectName
	self objectName: anObjectName.
	bytes := aStream contents.! !

!GitBlob methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 15:36'!
saveTo: aStream
	aStream nextPutAll: bytes.! !

!GitObject class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 07:16'!
computeTypecodeMap
	| m code |
	m := Dictionary new.
	self allSubclassesDo: [:c |
		code := c typeCode.
		code ifNotNil: [m at: code put: c]].
	^ m! !

!GitObject class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 07:15'!
initialize
	"GitObject initialize"
	super initialize.
	TypecodeMap := self computeTypecodeMap.! !

!GitObject class methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:44'!
loadFrom: aStream name: anObjectName
	| code length limitedStream |
	code := (aStream upTo: 32) asString.
	length := (aStream upTo: 0) asString asInteger.
	limitedStream := ReadStream on: (aStream next: length).
	^ (TypecodeMap at: code) new loadFrom: limitedStream name: anObjectName! !

!GitObject class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 17:04'!
typeCode
	"Override in subclasses, returning non-nil where you want them to be constructible from a git oject store."
	^ nil! !

!GitObject methodsFor: 'as yet unclassified' stamp: 'tonyg 7/3/2010 22:45'!
isStorable
	"All GitObjects are storable, except for empty GitTrees. There's no particular reason
	why empty GitTrees can't be stored, except that canonical git(1) doesn't like them.
	Running 'git fsck' on a repository with empty GitTrees in it produces a complaint, and
	cloning or manipulating such a repository causes mysterious empty commits which
	have the effect of fixing the problem for you."
	^ true! !

!GitObject methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:54'!
loadFrom: aStream name: anObjectName
	self subclassResponsibility! !

!GitObject methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:49'!
objectName

	^ objectName! !

!GitObject methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:49'!
objectName: anObject

	objectName := anObject.! !

!GitObject methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 23:04'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	aStream nextPutAll: (objectName ifNil: ['???'] ifNotNil: [:o | o resolve]).
	aStream nextPut: $).! !

!GitObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 15:35'!
saveTo: aStream
	self subclassResponsibility! !

!GitObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 15:47'!
typeCode
	^ self class typeCode! !

GitObject subclass: #GitTaglikeObject
	instanceVariableNames: 'message properties'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

GitTaglikeObject subclass: #GitCommit
	instanceVariableNames: 'tree parents author committer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitCommit class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 22:21'!
typeCode
	^ 'commit'! !

!GitCommit class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 22:06'!
withDefaultStamps
	| s |
	s := GitStamp now.
	^ self new author: s; committer: s! !

!GitCommit methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:54'!
author
	^ author! !

!GitCommit methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:54'!
author: anObject
	author := anObject! !

!GitCommit methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:54'!
committer
	^ committer! !

!GitCommit methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:54'!
committer: anObject
	committer := anObject! !

!GitCommit methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:23'!
loadKey: k value: v
	k = 'tree' ifTrue: [tree := GitObjectNameLiteral sha1String: v. ^ self].
	k = 'parent' ifTrue: [
		parents ifNil: [parents := OrderedCollection new].
		parents add: (GitObjectNameLiteral sha1String: v).
		^ self].
	k = 'author' ifTrue: [author := GitStamp fromString: v. ^ self].
	k = 'committer' ifTrue: [committer := GitStamp fromString: v. ^ self].
	"Otherwise it's treated generically."
	^ super loadKey: k value: v! !

!GitCommit methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 21:54'!
parents
	^ parents ifNil: [#()]! !

!GitCommit methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:54'!
parents: anObject
	^ parents := anObject! !

!GitCommit methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:27'!
saveFixedKeysTo: aStream
	"Looking at builtin-commit-tree.c, the order of properties is fixed:
	tree, parents, author, committer, everything else, message."
	self saveKey: 'tree' value: tree sha1String to: aStream.
	self parents do: [:p | self saveKey: 'parent' value: p sha1String to: aStream].
	self saveKey: 'author' value: author printString to: aStream.
	self saveKey: 'committer' value: committer printString to: aStream.! !

!GitCommit methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 21:53'!
tree
	^ tree! !

!GitCommit methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:54'!
tree: anObject
	tree := anObject! !

!GitCommit methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:34'!
validate
	tree ifNil: [self error: 'Cannot have a commit without a tree.'].
	author ifNil: [self error: 'Author identity is required.'].
	committer ifNil: [self error: 'Committer identity is required.'].
	super validate.! !

GitTaglikeObject subclass: #GitTag
	instanceVariableNames: 'targetName targetTypeCode tagName tagger'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTag class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:39'!
named: tagName on: objectName typeCode: typeCode message: message
	^ self named: tagName on: objectName typeCode: typeCode message: message tagger: GitStamp now! !

!GitTag class methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:51'!
named: tagName on: objectName typeCode: typeCode message: message tagger: aGitStamp
	^ self new
		tagName: tagName;
		targetName: objectName;
		targetTypeCode: typeCode;
		message: message;
		tagger: aGitStamp! !

!GitTag class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:16'!
typeCode
	^ 'tag'! !

!GitTag methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:52'!
loadKey: k value: v
	k = 'object' ifTrue: [targetName := GitObjectNameLiteral sha1String: v. ^ self].
	k = 'type' ifTrue: [targetTypeCode := v. ^ self].
	k = 'tag' ifTrue: [tagName := v. ^ self].
	k = 'tagger' ifTrue: [tagger := GitStamp fromString: v. ^ self].
	"Otherwise it's treated generically."
	^ super loadKey: k value: v! !

!GitTag methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:53'!
saveFixedKeysTo: aStream
	"From builtin-mktag.c:
	A signature file has a very simple fixed format: four lines
	of ""object <sha1>"" + ""type <typename>"" + ""tag <tagname>"" +
	""tagger <committer>"", followed by a blank line, a free-form tag
	message and a signature block that git itself doesn't care about,
	but that can be verified with gpg or similar."
	self saveKey: 'object' value: targetName sha1String to: aStream.
	self saveKey: 'type' value: (targetTypeCode ifNil: [targetName retrieveObject typeCode]) to: aStream.
	self saveKey: 'tag' value: tagName to: aStream.
	self saveKey: 'tagger' value: tagger printString to: aStream.! !

!GitTag methodsFor: 'accessing' stamp: 'tonyg 2/25/2010 16:30'!
tagName
	^ tagName! !

!GitTag methodsFor: 'accessing' stamp: 'tonyg 2/25/2010 16:30'!
tagName: anObject
	tagName := anObject! !

!GitTag methodsFor: 'accessing' stamp: 'tonyg 2/25/2010 16:30'!
tagger
	^ tagger! !

!GitTag methodsFor: 'accessing' stamp: 'tonyg 2/25/2010 16:30'!
tagger: anObject
	tagger := anObject! !

!GitTag methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:51'!
targetName
	^ targetName! !

!GitTag methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:51'!
targetName: anObject
	targetName := anObject! !

!GitTag methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:52'!
targetTypeCode
	^ targetTypeCode! !

!GitTag methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:52'!
targetTypeCode: anObject
	targetTypeCode := anObject! !

!GitTag methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:53'!
validate
	targetName ifNil: [self error: 'Cannot have a tag without a tagged object.'].
	tagName ifNil: [self error: 'Tag name is required.'].
	tagger ifNil: [self error: 'Tagger identity is required.'].
	super validate.! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:49'!
loadFrom: aStream name: anObjectName
	self objectName: anObjectName.
	[aStream peek = 10 "Character lf asInteger"] whileFalse: [
		| k v |
		k := (aStream upTo: 32 "Character space asInteger") asString.
		v := (aStream upTo: 10 "Character lf asInteger") asString.
		self loadKey: k value: v].
	aStream next.
	message := aStream upToEnd asString.
! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:23'!
loadKey: k value: v
	self properties add: k -> v.! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:22'!
message
	^ message! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:22'!
message: aString
	message := aString! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:22'!
properties
	^ properties ifNil: [properties := OrderedCollection new]! !

!GitTaglikeObject methodsFor: 'accessing' stamp: 'tonyg 2/25/2010 16:23'!
properties: anObject
	^ properties := anObject! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:27'!
saveFixedKeysTo: aStream
	self subclassResponsibility! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:24'!
saveKey: k value: v to: aStream
	aStream nextPutAll: k asByteArray; nextPut: 32; nextPutAll: v asByteArray; nextPut: 10.! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:27'!
saveTo: aStream
	self validate.
	self saveFixedKeysTo: aStream.
	properties ifNotNil: [properties do: [:assoc |
			self saveKey: assoc key value: assoc value to: aStream]].
	aStream nextPut: 10.
	aStream nextPutAll: message asByteArray.! !

!GitTaglikeObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:34'!
validate
	"Override in subclasses to provide pre-save validation checks.
	Don't forget to call super validate, though!!"
	message ifNil: [self error: 'Message is required.'].! !

GitObject subclass: #GitTree
	instanceVariableNames: 'entries'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTree class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:15'!
entries: aCollection
	| t |
	t := self new.
	aCollection do: [:each | t putEntry: each].
	^ t! !

!GitTree class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 07:17'!
typeCode
	^ 'tree'! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:10'!
at: aName
	^ (self entryAt: aName) object! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 10:02'!
diffFrom: baseTree
	^ (GitTreeDiffBuilder new initializeOld: baseTree new: self) build; result! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:15'!
entries
	^ entries! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 12:02'!
entriesByObjectName
	| d |
	d := Dictionary new.
	entries valuesDo: [:v | d at: v objectName put: v].
	^ d! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:10'!
entryAt: aName
	^ entries at: aName! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 20:18'!
initialize
	super initialize.
	entries := Dictionary new.! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 7/3/2010 22:46'!
isStorable
	"See the comment for GitObject>>isStorable."
	^ entries isEmpty not! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:53'!
loadFrom: aStream name: anObjectName
	self objectName: anObjectName.
	[aStream atEnd] whileFalse: [| e |
		e := GitTreeEntry new loadFrom: aStream.
		self putEntry: e]! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:11'!
putEntry: anEntry
	entries at: anEntry fileName put: anEntry! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:14'!
removeEntryAt: aFileName
	entries removeKey: aFileName! !

!GitTree methodsFor: 'as yet unclassified' stamp: 'tonyg 7/3/2010 22:27'!
saveTo: aStream
	entries values sorted do: [:e | e saveTo: aStream].! !

Object subclass: #GitObjectName
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitObjectName methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 06:45'!
resolve
	self subclassResponsibility! !

!GitObjectName methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 14:31'!
retrieveObject
	^ GitRepositoryCache instance objectAt: self! !

GitObjectName subclass: #GitObjectNameLiteral
	instanceVariableNames: 'sha1String'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitObjectNameLiteral class methodsFor: 'as yet unclassified' stamp: 'tonyg 1/20/2020 16:28'!
loadFrom: aStream
	^ self loadFromInteger: (aStream next: 20) asUnsignedBigEndianInteger! !

!GitObjectNameLiteral class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 16:05'!
loadFromInteger: h
	^ self sha1String: ((h printStringBase: 16) asLowercase forceTo: 40 paddingStartWith: $0)! !

!GitObjectNameLiteral class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 06:45'!
sha1String: aString
	^ self new sha1String: aString! !

!GitObjectNameLiteral methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 11:44'!
= other
	^ (self species = other species) and: [self sha1String = other sha1String]! !

!GitObjectNameLiteral methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 11:44'!
hash
	^ self sha1String hash! !

!GitObjectNameLiteral methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 07:09'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(; nextPutAll: sha1String; nextPut: $).! !

!GitObjectNameLiteral methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 06:45'!
resolve
	^ sha1String! !

!GitObjectNameLiteral methodsFor: 'accessing' stamp: 'tonyg 2/24/2010 01:03'!
saveTo: aStream
	| h dl |
	h := Integer readFrom: (ReadStream on: sha1String asUppercase) base: 16.
	dl := h digitLength.
	20 to: dl + 1 by: -1 do: [:digitIndex | aStream nextPut: 0].
	dl to: 1 by: -1 do: [:digitIndex | aStream nextPut: (h digitAt: digitIndex)].
! !

!GitObjectNameLiteral methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 06:44'!
sha1String
	^ sha1String! !

!GitObjectNameLiteral methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 06:44'!
sha1String: anObject
	sha1String := anObject! !

GitObjectName subclass: #GitObjectNameRef
	instanceVariableNames: 'repo name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitObjectNameRef class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 13:01'!
repo: r name: n
	^ self new initializeRepo: r name: n! !

!GitObjectNameRef methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 11:45'!
= other
	^ (self species = other species) and: [self repo == other repo and: [self refName = other refName]]! !

!GitObjectNameRef methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 11:46'!
hash
	^ self repo hash bitXor: self refName hash! !

!GitObjectNameRef methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 13:01'!
initializeRepo: r name: n
	repo := r.
	name := n.! !

!GitObjectNameRef methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 13:02'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(; nextPutAll: repo printString; nextPutAll: ', '; nextPutAll: name; nextPut: $).! !

!GitObjectNameRef methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 11:45'!
refName
	^ name! !

!GitObjectNameRef methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 11:45'!
repo
	^ repo! !

!GitObjectNameRef methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 16:02'!
resolve
	^ (repo resolveRef: name)
		ifNil: [(KeyNotFound key: self) signal]
		ifNotNil: [:n | n sha1String]! !

Object subclass: #GitPack
	instanceVariableNames: 'idx basePath fanoutTable objectCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitPack methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 15:42'!
basePath
	^ basePath! !

!GitPack methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 22:31'!
basePath: aString
	basePath := aString.
	idx := FileStream readOnlyFileNamed: basePath, '.idx'.
	idx binary.
	((idx next: 8) = #(255 116 79 99 0 0 0 2) asByteArray)
		ifFalse: [self error: 'Bad pack index header'].
	fanoutTable := idx next: 1024. "256 4-byte big-endian entries."
	objectCount := self fanoutTableAt: 255.! !

!GitPack methodsFor: 'accessing' stamp: 'tonyg 1/20/2020 16:31'!
crcAt: index
	"1024 fanout, 8 header, minus 4 for smalltalk's 1-based indexing"
	idx position: 1028 + (objectCount * 20) + (index * 4).
	^ (idx next: 4) asUnsignedBigEndianInteger.! !

!GitPack methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 22:31'!
fanoutTableAt: fanoutIndex
	^ fanoutTable unsignedLongAt: (fanoutIndex * 4) + 1 bigEndian: true.! !

!GitPack methodsFor: 'accessing' stamp: 'tonyg 2/24/2010 00:32'!
objectAt: sha1String
	| p |
	p := self packedObjectAt: sha1String.
	p ifNotNil: [^ p object].
	^ nil! !

!GitPack methodsFor: 'accessing' stamp: 'tonyg 1/20/2020 16:31'!
offsetAt: index
	| o |
	"1024 fanout, 8 header. We're still using zero-based indexing here."
	idx position: 1032 + (objectCount * 24) + (index * 4).
	o := (idx next: 4) asUnsignedBigEndianInteger.
	(o >= 16r80000000)
		ifFalse: [^ o]
		ifTrue: [
			"the table offset is zero-based here because it comes from git, not smalltalk"
			idx position: 1032 + (objectCount * 28) + ((o bitAnd: 16r7FFFFFFF) * 8).
			^ (idx next: 8) asUnsignedBigEndianInteger]! !

!GitPack methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 23:00'!
packedObjectAt: sha1String
	| i o |
	i := self tableIndexForSha: sha1String.
	i ifNil: [^ nil].
	o := self offsetAt: i.
	^ self packedObjectAtOffset: o name: (GitObjectNameLiteral sha1String: sha1String)! !

!GitPack methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:59'!
packedObjectAtOffset: offset name: name
	| p c |
	p := FileStream readOnlyFileNamed: basePath , '.pack'.
	p binary.
	p position: offset.
	c := GitPackedObject new pack: self; objectName: name.
	c loadFrom: p.
	^ c! !

!GitPack methodsFor: 'accessing' stamp: 'tonyg 1/20/2020 16:26'!
tableIndexForSha: s
	| fanoutIndex sNum index low high test |
	sNum := Integer readFrom: (ReadStream on: s asUppercase) base: 16.
	fanoutIndex := sNum bitShift: -152. "160 bit hash. We want the upper 8 bits."

	"We are searching using a zero-based index, since position: on the file is zero-based."
	low := (fanoutIndex > 0) ifTrue: [self fanoutTableAt: fanoutIndex - 1] ifFalse: [0].
	high := (self fanoutTableAt: fanoutIndex) - 1. "binary-search wants an inclusive range"

	[index := high + low // 2.
	low > high] whileFalse:[
		idx position: index * 20 + 1032. "256 * 4 for the fanoutTable, plus 8 for the header"
		test := (idx next: 20) asUnsignedBigEndianInteger.
		test = sNum 
			ifTrue:[^index]
			ifFalse:[sNum > test
				ifTrue: [low := index + 1]
				ifFalse: [high := index - 1]]].
	^ nil! !

Object subclass: #GitPackedObject
	instanceVariableNames: 'pack file decodeClass length headerPos unpackedBytes objectName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 20:09'!
applyDelta: delta to: base
	decodeClass := base decodeClass.
	unpackedBytes := delta runAgainst: base unpackedBytes.! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 20:09'!
decodeClass
	^ decodeClass! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 20:04'!
loadFrom: aStream
	| b type shift |
	file := aStream.
	headerPos := file position.
	b := file next.
	type := (b bitShift: -4) bitAnd: 7.
	length := b bitAnd: 15.
	shift := 4.
	[(b bitAnd: 128) isZero] whileFalse: [
		b := file next.
		length := length bitOr: ((b bitAnd: 127) bitShift: shift).
		shift := shift + 7].
	self unpackContentType: type.! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 22:46'!
object
	^ decodeClass new loadFrom: (ReadStream on: unpackedBytes) name: objectName! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 23:01'!
objectName: anObjectName
	objectName := anObjectName! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 11:05'!
pack
	^ pack! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 11:05'!
pack: anObject
	pack := anObject! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 20:07'!
readDelta
	^ GitDeltaInterpreter new instructions: (ZLibReadStream on: file) contents.! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 20:04'!
unpackContentType: type
	type = 1 "OBJ_COMMIT" ifTrue: [decodeClass := GitCommit. ^ self unpackSimpleContent].
	type = 2 "OBJ_TREE" ifTrue: [decodeClass := GitTree. ^ self unpackSimpleContent].
	type = 3 "OBJ_BLOB" ifTrue: [decodeClass := GitBlob. ^ self unpackSimpleContent].
	type = 4 "OBJ_TAG" ifTrue: [self error: 'Packed tags aren''t supported yet.'].
	type = 6 "OBJ_OFS_DELTA" ifTrue: [^ self unpackOfsDeltaContent].
	type = 7 "OBJ_REF_DELTA" ifTrue: [^ self unpackRefDeltaContent].
	self error: 'Invalid pack object type ', type printString.! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 1/27/2020 23:01'!
unpackOfsDeltaContent
	| offset b base delta |
	b := file next.
	offset := b bitAnd: 127.
	[(b bitAnd: 128) isZero] whileFalse: [
		b := file next.
		"This is bizarre, but it's what pack-format.txt and the Git source says is happening."
		offset := ((offset + 1) bitShift: 7) bitOr: (b bitAnd: 127)].

	delta := self readDelta.
	base := pack packedObjectAtOffset: headerPos - offset name: objectName.
	self applyDelta: delta to: base.! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 20:08'!
unpackRefDeltaContent
	| s delta base |
	s := (GitObjectNameLiteral loadFrom: file) resolve.
	delta := self readDelta.
	base := pack packedObjectAt: s.
	self applyDelta: delta to: base.! !

!GitPackedObject methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 20:05'!
unpackSimpleContent 
	unpackedBytes := (ZLibReadStream on: file) contents.! !

!GitPackedObject methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 20:09'!
unpackedBytes
	^ unpackedBytes! !

Object subclass: #GitRepository
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

GitRepository subclass: #GitDiskRepository
	instanceVariableNames: 'path'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/1/2020 21:04'!
config
	^ GitConfig new readFrom: ((self directory oldFileOrNoneNamed: 'config') ifNil: ['' readStream])! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 2/26/2010 10:59'!
createSkeleton
	"This is 'git init'. I'd call it 'initialize', but that has another use in Smalltalk ;-)"
	| d |
	d := self directory.
	d assureExistence.
	#(
		'objects'
		'objects/pack'
		'objects/info'
		'refs'
		'refs/heads'
		'refs/tags'
		'info'
	) do: [:each | (d directoryNamed: each) assureExistence].
	(d fileExists: 'HEAD') ifFalse: [
		(d newFileNamed: 'HEAD') nextPutAll: 'ref: refs/heads/master', String lf; close].
! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 06:55'!
directory
	^ FileDirectory on: path! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 10:55'!
looseObjectNames
	| r |
	r := OrderedCollection new.
	self looseObjectNamesDo: [ :n | r add: n ].
	^ r! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 10:55'!
looseObjectNamesDo: aBlock 
	| root |
	root := self directory directoryNamed: 'objects'.
	root directoryNames do: 
		[ :dirName | 
		| dir |
		dirName size = 2 ifTrue: 
			[ dir := root directoryNamed: dirName.
			dir fileNames do: [ :fileName | aBlock value: (GitObjectNameLiteral sha1String: dirName , fileName) ] ] ]! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 10:55'!
looseObjectPathFor: name 
	| s |
	s := name resolve.
	^ 'objects/' , (s first: 2) , '/' , (s allButFirst: 2)! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:46'!
looseObjectStreamAt: name 
	| deflatedFile |
	deflatedFile := self directory readOnlyFileNamed: (self looseObjectPathFor: name).
	^ ZLibReadStream on: deflatedFile binary! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 2/25/2010 16:45'!
makeTag: aGitTag
	^ self makeTagNamed: aGitTag tagName on: (self storeObject: aGitTag)! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 2/25/2010 16:45'!
makeTagNamed: tagName on: objectName
	^ self updateRef: 'refs/tags/', tagName to: objectName! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 22:44'!
objectAt: name 
	^ [ GitObject loadFrom: (self looseObjectStreamAt: name) name: name ] 
		on: FileDoesNotExistException
		do: [ :ex | ex return: (self packedObjectAt: name) ]! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 10:35'!
packedObjectAt: name
	| s o |
	s := name resolve.
	self packsDo: [:p |
		o := p objectAt: s.
		o ifNotNil: [^ o]].
	^ nil! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 16:47'!
packedRefs
	| f d |
	f := self directory oldFileOrNoneNamed: 'packed-refs'.
	f ifNil: [^ nil].
	
	d := Dictionary new.
	[f atEnd] whileFalse: [
		(f peek = $# or: [f peek = $^]) "Ignore comments and 'peeled' tags for now"
			ifTrue: [f upTo: Character lf]
			ifFalse: [| s n |
				s := f upTo: Character space.
				n := f upTo: Character lf.
				d at: n put: (GitObjectNameLiteral sha1String: s)]].
	f close.
	^ d! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 16:06'!
packsDo: aBlock
	| d |
	d := self directory directoryNamed: 'objects/pack'.
	d fileNames do: [:n |
		(n endsWith: '.idx') ifTrue: [
			n := n allButLast: 4.
			aBlock value: (GitPack new basePath: (d fullNameFor: n))]]! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 06:42'!
path
	^ path! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 06:54'!
path: aString
	path := aString! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/1/2020 20:44'!
refs
	| d fs |
	d := self packedRefs.

	self directory fileEntries
		select: [:e | e name endsWith: 'HEAD']
		thenDo: [:e | [d at: e name put: (self resolveLooseRef: e readStream)]
			on: FileDoesNotExistException do: [:ex | ]].

	fs := OrderedCollection new.
	(self directory directoryNamed: 'refs') directoryTreeDo: [:i | fs add: i copy].
	fs select: [:e | e last isDirectory not] thenDo: [:e |
		[d
			at: ((e collect: [:x | x name]) joinSeparatedBy: '/')
			put: (self resolveLooseRef: e last readStream)]
		on: FileDoesNotExistException do: [:ex | ]].
	^ d! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/1/2020 20:40'!
resolveLooseRef: refFileStream
	| target |
	target := refFileStream contents.
	(target beginsWith: 'ref:') ifTrue: [^ self resolveRef: (target allButFirst: 4) withBlanksTrimmed].
	^ GitObjectNameLiteral sha1String: target substrings first "Ignore other junk after the hash."! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/1/2020 20:41'!
resolveRef: name
"From the git-rev-parse manpage:
            1. if $GIT_DIR/<name> exists, that is what you mean (this is usually useful only for HEAD, FETCH_HEAD, ORIG_HEAD and MERGE_HEAD);
            2. otherwise, $GIT_DIR/refs/<name> if exists;
            3. otherwise, $GIT_DIR/refs/tags/<name> if exists;
            4. otherwise, $GIT_DIR/refs/heads/<name> if exists;
            5. otherwise, $GIT_DIR/refs/remotes/<name> if exists;
            6. otherwise, $GIT_DIR/refs/remotes/<name>/HEAD if exists."
	| fullName havePackedRefs packedRefs result |
	havePackedRefs := false.
	packedRefs := nil.
	#(
		''
		'refs/'
		'refs/tags/'
		'refs/heads/'
		'refs/remotes/'
		"We don't do the last case (number 6) from the manual yet."
	) do: [:prefix |
		fullName := prefix, name.
		result := [self resolveLooseRef: (self directory readOnlyFileNamed: fullName)]
			on: FileDoesNotExistException do: [:ex | nil].
		result ifNotNil: [^ result].
		havePackedRefs ifFalse: [packedRefs := self packedRefs. havePackedRefs := true].
		packedRefs ifNotNil: [
			result := packedRefs at: fullName ifAbsent: [nil].
			result ifNotNil: [^ result]]].
	^ nil! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 7/3/2010 22:46'!
storeLooseObject: o
	| s m body objectBytes name |
	o isStorable ifFalse: [^ nil].
	s := WriteStream on: (ByteArray new).
	o saveTo: s.
	body := s contents.
	
	m := WriteStream on: (ByteArray new).
	m
		nextPutAll: o typeCode asByteArray; nextPut: 32;
		nextPutAll: body size printString asByteArray; nextPut: 0;
		nextPutAll: body.
	objectBytes := m contents.
	
	name := GitObjectNameLiteral loadFromInteger:
				(SecureHashAlgorithm new hashMessage: objectBytes).
	self storeLooseObjectBytes: objectBytes at: name.
	^ name! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 16:46'!
storeLooseObjectBytes: aByteArray at: name
	| z compressed |
	z := ZLibWriteStream on: (ByteArray new).
	z nextPutAll: aByteArray; close.
	compressed := z encodedStream contents.
	FileDirectory splitName: (self looseObjectPathFor: name) to: [:oPath :oFile |
		[((self directory directoryNamed: oPath) assureExistence; newFileNamed: oFile)
			binary; nextPutAll: compressed; close]
				on: FileExistsException do: [:ex |
					"Leave existing files alone. We could cross-check their contents, if we liked."
					ex return: nil]].! !

!GitDiskRepository methodsFor: 'accessing' stamp: 'tonyg 2/25/2010 16:43'!
storeObject: o
	^ self storeLooseObject: o! !

!GitDiskRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/1/2020 20:55'!
updateRef: refPath to: objectName
	| target |
	target := (self directory oldFileOrNoneNamed: refPath) ifNotNil: [:f | f contents].

	(target isNil not and: [target beginsWith: 'ref:']) ifTrue: [
		^ self updateRef: (target allButFirst: 4) withBlanksTrimmed to: objectName].

	FileDirectory splitName: refPath to: [:rPath :rFile |
		((self directory directoryNamed: rPath) assureExistence; fileNamed: rFile)
			truncate; nextPutAll: objectName resolve; nextPut: Character lf; close].

	^ objectName! !

!GitRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/26/2010 11:00'!
createSkeleton
	self subclassResponsibility! !

!GitRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/26/2010 11:00'!
makeTag: aGitTag
	self subclassResponsibility! !

!GitRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/26/2010 11:01'!
makeTagNamed: tagName on: objectName
	self subclassResponsibility! !

!GitRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/26/2010 11:01'!
objectAt: name
	self subclassResponsibility! !

!GitRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/26/2010 11:01'!
resolveRef: name
	self subclassResponsibility! !

!GitRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/26/2010 11:02'!
storeObject: o
	self subclassResponsibility! !

!GitRepository methodsFor: 'as yet unclassified' stamp: 'tonyg 2/26/2010 11:02'!
updateRef: refPath to: objectName
	self subclassResponsibility! !

Object subclass: #GitRepositoryCache
	instanceVariableNames: 'repos mutex'
	classVariableNames: 'Instance'
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitRepositoryCache class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 08:17'!
instance
	^ Instance ifNil: [Instance := self new]! !

!GitRepositoryCache methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 14:26'!
addRepository: repo
	mutex critical: [
		(repos includes: repo)
			ifFalse: [repos add: repo]].
	self changed: (Message selector: #add: argument: repo).! !

!GitRepositoryCache methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 14:38'!
initialize
	super initialize.
	repos := OrderedCollection new.
	mutex := Mutex new.! !

!GitRepositoryCache methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 10:53'!
objectAt: objectName
	| resolved i held o |
	mutex critical: [
		repos isEmpty ifTrue: [^ nil].
		resolved := GitObjectNameLiteral sha1String: objectName resolve.
		i := 1.
		held := repos at: 1.
		[i > repos size] whileFalse: [
			| repo |
			repo := repos at: i.
			repos at: i put: held.
			o := repo objectAt: resolved.
			o ifNotNil: [
				repos at: 1 put: repo.
				^ o].
			i := i + 1.
			held := repo.].
		repos at: 1 put: held.
		^ nil.]! !

!GitRepositoryCache methodsFor: 'as yet unclassified' stamp: 'tonyg 1/27/2020 14:26'!
removeRepository: repo
	mutex critical: [
		repos remove: repo ifAbsent: []].
	self changed: (Message selector: #remove: argument: repo).! !

!GitRepositoryCache methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 14:39'!
repos
	mutex critical: [
		^ repos copy]! !

Object subclass: #GitStamp
	instanceVariableNames: 'name email timeStamp'
	classVariableNames: 'DefaultEmail DefaultName'
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitStamp class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 22:15'!
defaultEmail
	^ DefaultEmail ifNil: [self requestDefaultEmail]! !

!GitStamp class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 22:15'!
defaultName
	^ DefaultName ifNil: [self requestDefaultName]! !

!GitStamp class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 21:29'!
fromString: aString
	^ self readFrom: (ReadStream on: aString)! !

!GitStamp class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 22:12'!
now
	"TODO: make timestamp now have the correct timezone!! Bug in Squeak, or am I overlooking something obvious?"
	^ self new name: self defaultName; email: self defaultEmail; timeStamp: TimeStamp now! !

!GitStamp class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 21:40'!
readFrom: aStream
	| n e s sign z |
	n := (aStream upTo: $<) withBlanksTrimmed.
	e := (aStream upTo: $>) withBlanksTrimmed.
	aStream skipSeparators.
	s := Integer readFrom: aStream.
	aStream skipSeparators.
	sign := aStream next.
	('-+' includes: sign) ifFalse: [self error: 'Invalid timezone sign'].
	z := Integer readFrom: aStream.
	z := (z // 100) hours + (z \\ 100) minutes.
	sign = $- ifTrue: [z := z negated].
	
	^ self new
		name: n;
		email: e;
		timeStamp: ((DateAndTime unixEpoch + s seconds) utcOffset: z)! !

!GitStamp class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 22:16'!
requestDefaultEmail
	^ DefaultEmail := FillInTheBlank request: 'Email address to use in Git commits by default'! !

!GitStamp class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 22:16'!
requestDefaultName
	^ DefaultName := FillInTheBlank request: 'Name to use in Git commits by default'! !

!GitStamp methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:24'!
cleanUp: aString
	^ aString copyWithoutAll: '<>', String lf! !

!GitStamp methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:21'!
email
	^ email! !

!GitStamp methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:23'!
email: aString
	email := self cleanUp: aString! !

!GitStamp methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:22'!
name
	^ name! !

!GitStamp methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:23'!
name: aString
	name := self cleanUp: aString! !

!GitStamp methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:43'!
printOn: aStream
	| z |
	aStream
		nextPutAll: name; nextPutAll: ' <'; nextPutAll: email; nextPutAll: '> ';
		nextPutAll: (timeStamp - DateAndTime unixEpoch) asSeconds printString;
		nextPut: Character space.
	z := timeStamp offset.
	z < Duration zero
		ifTrue: [aStream nextPut: $-. z := z negated]
		ifFalse: [aStream nextPut: $+].
	aStream
		nextPutAll: (z hours printString forceTo: 2 paddingWith: $0);
		nextPutAll: (z minutes printString forceTo: 2 paddingWith: $0).! !

!GitStamp methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:21'!
timeStamp
	^ timeStamp! !

!GitStamp methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 21:23'!
timeStamp: aDuration
	timeStamp := aDuration! !

Object subclass: #GitTreeDiffBuilder
	instanceVariableNames: 'old new result path'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTreeDiffBuilder methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 12:05'!
addEntry: item
	result at: path asArray put: item! !

!GitTreeDiffBuilder methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 10:06'!
build
	self visitTreeOld: old new: new.! !

!GitTreeDiffBuilder methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 10:27'!
initializeOld: oldTree new: newTree 
	old := oldTree.
	new := newTree.
	result := Dictionary new.
	path := OrderedCollection new.! !

!GitTreeDiffBuilder methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 12:06'!
result
	^ result! !

!GitTreeDiffBuilder methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 12:09'!
visitOld: oldItem new: newItem
	| oldType newType |
	oldType := oldItem object typeCode.
	newType := newItem object typeCode.
	
	oldType = newType ifFalse: [
		^ self addEntry: (GitTreeDiffTypeChangedItem new oldItem: oldItem; newItem: newItem; yourself)].
	
	oldItem mode = newItem mode ifFalse: [
		self addEntry: (GitTreeDiffModeChangedItem new oldItem: oldItem; newItem: newItem; yourself)].

	oldType = 'tree' ifTrue: [
		^ self visitTreeOld: oldItem object new: newItem object].
	oldType = 'blob' ifTrue: [
		^ self addEntry: (GitTreeDiffChangedItem new oldItem: oldItem; newItem: newItem; yourself)].
	self error: 'Unsupported item type ', oldType! !

!GitTreeDiffBuilder methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 12:11'!
visitTreeOld: o new: n
	| oldKeys newKeys oldNameMap newNameMap moved |
	oldKeys := o entries keys asSet.
	newKeys := n entries keys asSet.
	
	moved := Set new.
	oldNameMap := o entriesByObjectName.
	newNameMap := n entriesByObjectName.
	(oldNameMap keys intersection: newNameMap keys) do: [:nm | | oi ni |
		oi := oldNameMap at: nm.
		ni := newNameMap at: nm.
		oi fileName = ni fileName ifFalse: [
			moved add: nm.
			self addEntry: (GitTreeDiffRenamedItem new oldItem: oi; newItem: ni; yourself)]].
	 
	(oldKeys difference: newKeys) do: [:k | | e |
		e := o entries at: k.
		(moved includes: e objectName) ifFalse: [
			self addEntry: (GitTreeDiffDeletedItem new item: e)]].
	(newKeys difference: oldKeys) do: [:k | | e |
		e := n entries at: k.
		(moved includes: e objectName) ifFalse: [
			self addEntry: (GitTreeDiffNewItem new item: e)]].

	(oldKeys intersection: newKeys) do: [:k | | oi ni |
		oi := o entries at: k.
		ni := n entries at: k.
		oi objectName = ni objectName ifFalse: [self withPath: k do: [self visitOld: oi new: ni]]].! !

!GitTreeDiffBuilder methodsFor: 'as yet unclassified' stamp: 'tonyg 1/29/2020 10:28'!
withPath: edgeLabel do: aBlock
	| v |
	path add: edgeLabel.
	v := aBlock value.
	path removeLast.
	^ v! !

Object subclass: #GitTreeDiffEntry
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

GitTreeDiffEntry subclass: #GitTreeDiffChangedItem
	instanceVariableNames: 'oldItem newItem diff'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTreeDiffChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 12:12'!
computeDiff
	[
		^ TextDiffBuilder from: (oldItem object bytes utf8Decoded) to: (newItem object bytes utf8Decoded)
	] on: InvalidUTF8 do: [:ex | ^ nil]! !

!GitTreeDiffChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:19'!
diff
	diff ifNil: [diff := self computeDiff].
	^ diff! !

!GitTreeDiffChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:09'!
newItem

	^ newItem! !

!GitTreeDiffChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:09'!
newItem: anObject

	newItem := anObject.! !

!GitTreeDiffChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:09'!
oldItem

	^ oldItem! !

!GitTreeDiffChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:09'!
oldItem: anObject

	oldItem := anObject.! !

GitTreeDiffEntry subclass: #GitTreeDiffDeletedItem
	instanceVariableNames: 'item'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTreeDiffDeletedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:08'!
item

	^ item! !

!GitTreeDiffDeletedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:08'!
item: anObject

	item := anObject.! !

GitTreeDiffEntry subclass: #GitTreeDiffModeChangedItem
	instanceVariableNames: 'oldItem newItem'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTreeDiffModeChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
newItem

	^ newItem! !

!GitTreeDiffModeChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
newItem: anObject

	newItem := anObject.! !

!GitTreeDiffModeChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
oldItem

	^ oldItem! !

!GitTreeDiffModeChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
oldItem: anObject

	oldItem := anObject.! !

GitTreeDiffEntry subclass: #GitTreeDiffNewItem
	instanceVariableNames: 'item'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTreeDiffNewItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:08'!
item

	^ item! !

!GitTreeDiffNewItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:08'!
item: anObject

	item := anObject.! !

GitTreeDiffEntry subclass: #GitTreeDiffRenamedItem
	instanceVariableNames: 'oldItem newItem'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTreeDiffRenamedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
newItem

	^ newItem! !

!GitTreeDiffRenamedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
newItem: anObject

	newItem := anObject.! !

!GitTreeDiffRenamedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
oldItem

	^ oldItem! !

!GitTreeDiffRenamedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
oldItem: anObject

	oldItem := anObject.! !

GitTreeDiffEntry subclass: #GitTreeDiffTypeChangedItem
	instanceVariableNames: 'oldItem newItem'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTreeDiffTypeChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
newItem

	^ newItem! !

!GitTreeDiffTypeChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
newItem: anObject

	newItem := anObject.! !

!GitTreeDiffTypeChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
oldItem

	^ oldItem! !

!GitTreeDiffTypeChangedItem methodsFor: 'accessing' stamp: 'tonyg 1/29/2020 10:46'!
oldItem: anObject

	oldItem := anObject.! !

Object subclass: #GitTreeEntry
	instanceVariableNames: 'fileName mode objectName'
	classVariableNames: 'TypecodeMap'
	poolDictionaries: ''
	category: 'Git-Plumbing'!

!GitTreeEntry class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 20:16'!
dirMode
	^ 8r040000! !

!GitTreeEntry class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 20:15'!
fileMode
	^ 8r100644! !

!GitTreeEntry class methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:17'!
fileName: aString mode: anInteger objectName: aName
	^ self new initializeFileName: aString mode: anInteger objectName: aName! !

!GitTreeEntry class methodsFor: 'as yet unclassified' stamp: 'tonyg 7/3/2010 21:25'!
symlinkMode
	^ 8r120000! !

!GitTreeEntry methodsFor: 'as yet unclassified' stamp: 'tonyg 7/3/2010 22:27'!
<= aGitTreeEntry
	^ fileName <= aGitTreeEntry fileName! !

!GitTreeEntry methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 08:08'!
fileName
	^ fileName! !

!GitTreeEntry methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 19:18'!
initializeFileName: aString mode: anInteger objectName: aName
	fileName := aString.
	mode := anInteger.
	objectName := aName.! !

!GitTreeEntry methodsFor: 'as yet unclassified' stamp: 'tonyg 2/22/2010 08:13'!
loadFrom: aStream
	mode := Integer readFrom: (ReadStream on: (aStream upTo: 32) asString) base: 8.
	fileName := (aStream upTo: 0) asString.
	objectName := GitObjectNameLiteral loadFrom: aStream.! !

!GitTreeEntry methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 08:08'!
mode
	^ mode! !

!GitTreeEntry methodsFor: 'accessing' stamp: 'tonyg 2/23/2010 13:28'!
object
	^ objectName retrieveObject! !

!GitTreeEntry methodsFor: 'accessing' stamp: 'tonyg 2/22/2010 08:08'!
objectName
	^ objectName! !

!GitTreeEntry methodsFor: 'as yet unclassified' stamp: 'tonyg 7/3/2010 22:47'!
saveTo: aStream.
	objectName ifNil: [^ self].
	aStream nextPutAll: (mode printStringBase: 8) asByteArray; nextPut: 32.
	aStream nextPutAll: fileName asByteArray; nextPut: 0.
	objectName saveTo: aStream.! !

TestCase subclass: #GitBasicTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Tests'!

!GitBasicTests methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 21:43'!
testStampReading
	| str stamp |
	str := 'Tony Garnock-Jones <tonygarnockjones@gmail.com> 1266614121 +1300'.
	stamp := GitStamp fromString: str.
	self assert: str = stamp printString.! !

TestCase subclass: #GitPlumbingTests
	instanceVariableNames: 'r'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Git-Tests'!

!GitPlumbingTests methodsFor: 'as yet unclassified' stamp: 'tonyg 2/26/2010 10:58'!
setUp
	r := GitDiskRepository new path: (FileDirectory default fullNameFor: 'gitTestRepository.git').
	r directory exists ifTrue: [ r directory recursiveDelete ].
	r createSkeleton! !

!GitPlumbingTests methodsFor: 'as yet unclassified' stamp: 'tonyg 2/25/2010 16:48'!
testInitialCommit
	| b t c |
	b := r storeObject: (GitBlob bytes: 'hello').
	t := r storeObject: (GitTree entries: {
		GitTreeEntry fileName: 'test' mode: GitTreeEntry fileMode objectName: b}).
	c := r storeObject: (GitCommit withDefaultStamps tree: t; message: 'First commit').
	r makeTag: (GitTag named: 'tagtest' on: c typeCode: GitCommit typeCode message: 'First tag').
	r makeTagNamed: 'lighttag' on: c.
	r updateRef: 'HEAD' to: c.! !

!GitPlumbingTests methodsFor: 'as yet unclassified' stamp: 'tonyg 2/23/2010 18:26'!
testSkeleton
	"effectively tests just setUp."! !

!ByteArray methodsFor: '*Git-Core-converting' stamp: 'tonyg 1/20/2020 16:26'!
asUnsignedBigEndianInteger
	| i n |
	n := self size.
	i := LargePositiveInteger new: n.
	i replaceFrom: 1 to: n with: self reversed startingAt: 1.
	^ i normalize! !
GitObject initialize!
